import json
import time
from abc import ABC
from datetime import datetime
from threading import Thread

import urllib3
from typing import TYPE_CHECKING, Dict, List, Optional, Set

from concurrent.futures import ThreadPoolExecutor, as_completed

from skyplane import exceptions
from skyplane.api.config import TransferConfig
from skyplane.chunk import ChunkState, Chunk
from skyplane.utils import logger, imports
from skyplane.utils.fn import do_parallel
from skyplane.api.usage import UsageClient
from skyplane.utils.definitions import GB

from skyplane.cli.impl.common import print_stats_completed

if TYPE_CHECKING:
    from skyplane.api.transfer_job import TransferJob


class TransferHook(ABC):
    """Hook that shows transfer related stats"""

    def on_dispatch_start(self):
        """Starting the dispatch job"""
        raise NotImplementedError()

    def on_chunk_dispatched(self, chunks: List[Chunk]):
        """Dispatching data chunks to transfer"""
        raise NotImplementedError()

    def on_dispatch_end(self):
        """Ending the dispatch job"""
        raise NotImplementedError()

    def on_chunk_completed(self, chunks: List[Chunk], region_tag: Optional[str] = None):
        """Chunks are all transferred"""
        raise NotImplementedError()

    def on_transfer_end(self, transfer_stats):
        """Ending the transfer job"""
        raise NotImplementedError()

    def on_transfer_error(self, error):
        """Showing the tranfer error if it fails"""
        raise NotImplementedError()


class EmptyTransferHook(TransferHook):
    """Empty transfer hook that does nothing"""

    def __init__(self):
        return

    def on_dispatch_start(self):
        return

    def on_chunk_dispatched(self, chunks: List[Chunk]):
        return

    def on_dispatch_end(self):
        return

    def on_chunk_completed(self, chunks: List[Chunk], region_tag: Optional[str] = None):
        return

    def on_transfer_end(self, transfer_stats):
        return

    def on_transfer_error(self, error):
        return


class TransferProgressTracker(Thread):
    """Tracks transfer progress in one tranfer session"""

    def __init__(self, dataplane, jobs: List["TransferJob"], transfer_config: TransferConfig, hooks: TransferHook):
        """
        :param dataplane: dataplane that starts the transfer
        :type dataplane: Dataplane
        :param jobs: list of transfer jobs launched in parallel
        :type jobs: List
        :param transfer_config: the configuration during the transfer
        :type transfer_config: TransferConfig
        :param hooks: the hook that shows transfer related stats
        :type hooks: TransferHook
        """
        super().__init__()
        self.dataplane = dataplane
        self.jobs = {job.uuid: job for job in jobs}
        self.transfer_config = transfer_config

        if hooks is None:
            self.hooks = EmptyTransferHook()
        else:
            self.hooks = hooks

        # log job details
        logger.fs.debug(f"[TransferProgressTracker] Using dataplane {dataplane}")
        logger.fs.debug(f"[TransferProgressTracker] Initialized with {len(jobs)} jobs:")
        for job_uuid, job in self.jobs.items():
            logger.fs.debug(f"[TransferProgressTracker]   * {job_uuid}: {job}")
        logger.fs.debug(f"[TransferProgressTracker] Transfer config: {transfer_config}")

        # transfer state
        self.job_chunk_requests: Dict[str, Dict[str, Chunk]] = {}
        self.job_pending_chunk_ids: Dict[str, Dict[str, Set[str]]] = {}
        self.job_complete_chunk_ids: Dict[str, Dict[str, Set[str]]] = {}
        self.errors: Optional[Dict[str, List[str]]] = None

        # http_pool
        self.http_pool = urllib3.PoolManager(retries=urllib3.Retry(total=9))

    def __str__(self):
        return f"TransferProgressTracker({self.dataplane}, {self.jobs})"

    def run(self):
        """Dispatch and start the transfer jobs"""
        src_cloud_provider = self.dataplane.topology.src_region_tag.split(":")[0]
        dst_cloud_provider = self.dataplane.topology.dest_region_tags[0].split(":")[0]
        args = {
            "cmd": ",".join([job.__class__.__name__ for job in self.jobs.values()]),
            "recursive": ",".join([str(job.recursive) for job in self.jobs.values()]),
            "multipart": self.transfer_config.multipart_enabled,
            # "instances_per_region": 1,  # TODO: read this from config file
            # "src_instance_type": getattr(self.transfer_config, f"{src_cloud_provider}_instance_class"),
            # "dst_instance_type": #getattr(self.transfer_config, f"{dst_cloud_provider}_instance_class"),
            # "src_spot_instance": getattr(self.transfer_config, f"{src_cloud_provider}_use_spot_instances"),
            # "dst_spot_instance": getattr(self.transfer_config, f"{dst_cloud_provider}_use_spot_instances"),
        }
        # TODO: eventually jobs should be able to be concurrently dispatched and executed
        # however this will require being able to handle conflicting multipart uploads ids

        # initialize everything first
        for job_uuid, job in self.jobs.items():
            self.job_chunk_requests[job_uuid] = {}
            self.job_pending_chunk_ids[job_uuid] = {region: set() for region in self.dataplane.topology.dest_region_tags}
            self.job_complete_chunk_ids[job_uuid] = {region: set() for region in self.dataplane.topology.dest_region_tags}

        session_start_timestamp_ms = int(time.time() * 1000)
        for job_uuid, job in self.jobs.items():
            # pre-dispatch chunks to begin pre-buffering chunks
            try:
                chunk_stream = job.dispatch(self.dataplane, transfer_config=self.transfer_config)
                logger.fs.debug(f"[TransferProgressTracker] Dispatching job {job.uuid}")
                for chunk in chunk_stream:
                    chunks_dispatched = [chunk]
                    # TODO: check chunk ID
                    self.job_chunk_requests[job_uuid][chunk.chunk_id] = chunk
                    assert job_uuid in self.job_chunk_requests and chunk.chunk_id in self.job_chunk_requests[job_uuid]
                    self.hooks.on_chunk_dispatched(chunks_dispatched)
                    for region in self.dataplane.topology.dest_region_tags:
                        self.job_pending_chunk_ids[job_uuid][region].add(chunk.chunk_id)

                logger.fs.debug(
                    f"[TransferProgressTracker] Job {job.uuid} dispatched with {len(self.job_chunk_requests[job_uuid])} chunk requests"
                )
            except Exception as e:
                UsageClient.log_exception(
                    "dispatch job",
                    e,
                    args,
                    self.dataplane.topology.src_region_tag,
                    self.dataplane.topology.dest_region_tags[0],  # TODO: support multiple destinations
                    session_start_timestamp_ms,
                )
                raise e

            self.hooks.on_dispatch_end()

            def monitor_single_dst_helper(dst_region):
                start_time = time.time()
                try:
                    self.monitor_transfer(dst_region)
                except exceptions.SkyplaneGatewayException as err:
                    reformat_err = Exception(err.pretty_print_str()[37:])
                    UsageClient.log_exception(
                        "monitor transfer",
                        reformat_err,
                        args,
                        self.dataplane.topology.src_region_tag,
                        dst_region,
                        session_start_timestamp_ms,
                    )
                    raise err
                except Exception as e:
                    UsageClient.log_exception(
                        "monitor transfer", e, args, self.dataplane.topology.src_region_tag, dst_region, session_start_timestamp_ms
                    )
                    raise e
                end_time = time.time()

                runtime_s = end_time - start_time
                # transfer successfully completed
                transfer_stats = {
                    "dst_region": dst_region,
                    "total_runtime_s": round(runtime_s, 4),
                }

            results = []
            dest_regions = self.dataplane.topology.dest_region_tags
            with ThreadPoolExecutor(max_workers=len(dest_regions)) as executor:
                e2e_start_time = time.time()
                try:
                    future_list = [executor.submit(monitor_single_dst_helper, dest) for dest in dest_regions]
                    for future in as_completed(future_list):
                        results.append(future.result())
                except Exception as e:
                    raise e
            e2e_end_time = time.time()
            transfer_stats = {
                "total_runtime_s": e2e_end_time - e2e_start_time,
                "throughput_gbits": self.query_bytes_dispatched() / (e2e_end_time - e2e_start_time) / GB * 8,
            }
            self.hooks.on_transfer_end()

            int(time.time())
            try:
                for job in self.jobs.values():
                    logger.fs.debug(f"[TransferProgressTracker] Finalizing job {job.uuid}")
                    job.finalize()
            except Exception as e:
                UsageClient.log_exception(
                    "finalize job",
                    e,
                    args,
                    self.dataplane.topology.src_region_tag,
                    self.dataplane.topology.dest_region_tags[0],
                    session_start_timestamp_ms,
                )
                raise e
            int(time.time())

            # verify transfer
            try:
                for job in self.jobs.values():
                    logger.fs.debug(f"[TransferProgressTracker] Verifying job {job.uuid}")
                    job.verify()
            except Exception as e:
                UsageClient.log_exception(
                    "verify job",
                    e,
                    args,
                    self.dataplane.topology.src_region_tag,
                    self.dataplane.topology.dest_region_tags[0],
                    session_start_timestamp_ms,
                )
                raise e

            # transfer successfully completed
            UsageClient.log_transfer(
                transfer_stats,
                args,
                self.dataplane.topology.src_region_tag,
                self.dataplane.topology.dest_region_tags,
                session_start_timestamp_ms,
            )
            print_stats_completed(total_runtime_s=transfer_stats["total_runtime_s"], throughput_gbits=transfer_stats["throughput_gbits"])

    @imports.inject("pandas")
    def monitor_transfer(pd, self, region_tag):
        """Monitor the tranfer by copying remote gateway logs and show transfer stats by hooks"""
        # todo implement transfer monitoring to update job_complete_chunk_ids and job_pending_chunk_ids while the transfer is in progress

        # regions that are sinks for specific region tag
        # TODO: should eventualy map bucket to list of instances
        sinks = [n for nodes in self.dataplane.topology.sink_instances(region_tag).values() for n in nodes]

        while any([len(self.job_pending_chunk_ids[job_uuid][region_tag]) > 0 for job_uuid in self.job_pending_chunk_ids]):
            # refresh shutdown status by running noop
            do_parallel(lambda i: i.run_command("echo 1"), self.dataplane.bound_nodes.values(), n=8)

            # check for errors and exit if there are any (while setting debug flags)
            errors = self.dataplane.check_error_logs()
            if any(errors.values()):
                logger.warning("Copying gateway logs...")
                self.dataplane.copy_gateway_logs()
                self.errors = errors
                raise exceptions.SkyplaneGatewayException("Transfer failed with errors", errors)

            log_df = pd.DataFrame(self._query_chunk_status())
            if log_df.empty:
                logger.warning("No chunk status log entries yet")
                time.sleep(0.05)
                continue

            # TODO: have visualization for completition across all destinations
            is_complete_rec = (
                lambda row: row["state"] == ChunkState.complete
                and row["instance"] in [s.gateway_id for s in sinks]
                # and row["region_tag"] in region_sinks
            )
            sink_status_df = log_df[log_df.apply(is_complete_rec, axis=1)]
            completed_chunk_ids = list(set(sink_status_df.chunk_id.unique()))

            # update job_complete_chunk_ids and job_pending_chunk_ids
            # TODO: do chunk-tracking per-destination
            for job_uuid, job in self.jobs.items():
                try:
                    job_complete_chunk_ids = set(
                        chunk_id for chunk_id in completed_chunk_ids if self._chunk_to_job_map[chunk_id] == job_uuid
                    )
                except Exception as e:
                    raise e
                new_chunk_ids = (
                    self.job_complete_chunk_ids[job_uuid][region_tag]
                    .union(job_complete_chunk_ids)
                    .difference(self.job_complete_chunk_ids[job_uuid][region_tag])
                )
                completed_chunks = []
                for id in new_chunk_ids:
                    assert (
                        job_uuid in self.job_chunk_requests and id in self.job_chunk_requests[job_uuid]
                    ), f"Missing chunk id {id} for job {job_uuid}: {self.job_chunk_requests}"
                for id in new_chunk_ids:
                    completed_chunks.append(self.job_chunk_requests[job_uuid][id])
                self.hooks.on_chunk_completed(completed_chunks, region_tag)
                self.job_complete_chunk_ids[job_uuid][region_tag] = self.job_complete_chunk_ids[job_uuid][region_tag].union(
                    job_complete_chunk_ids
                )
                self.job_pending_chunk_ids[job_uuid][region_tag] = self.job_pending_chunk_ids[job_uuid][region_tag].difference(
                    job_complete_chunk_ids
                )
            # sleep
            time.sleep(0.05)

    @property
    # TODO: this is a very slow function, but we can't cache it since self.job_chunk_requests changes over time
    # do not call it more often than necessary
    def _chunk_to_job_map(self):
        return {chunk_id: job_uuid for job_uuid, cr_dict in self.job_chunk_requests.items() for chunk_id in cr_dict.keys()}

    def _query_chunk_status(self):
        def get_chunk_status(args):
            node, instance = args
            reply = self.http_pool.request("GET", f"{instance.gateway_api_url}/api/v1/chunk_status_log")
            if reply.status != 200:
                raise Exception(
                    f"Failed to get chunk status from gateway instance {instance.instance_name()}: {reply.data.decode('utf-8')}"
                )
            logs = []
            for log_entry in json.loads(reply.data.decode("utf-8"))["chunk_status_log"]:
                log_entry["region_tag"] = node.region_tag
                log_entry["instance"] = node.gateway_id
                log_entry["time"] = datetime.fromisoformat(log_entry["time"])
                log_entry["state"] = ChunkState.from_str(log_entry["state"])
                logs.append(log_entry)

            return logs

        rows = []
        for result in do_parallel(get_chunk_status, self.dataplane.bound_nodes.items(), n=8, return_args=False):
            rows.extend(result)
        return rows

    @property
    def is_complete(self, region_tag: str):
        """Return if the transfer is complete"""
        return all([len(self.job_pending_chunk_ids[job_uuid][region_tag]) == 0 for job_uuid in self.jobs.keys()])

    def query_bytes_remaining(self, region_tag: Optional[str] = None):
        """Query the total number of bytes remaining in all the transfer jobs"""
        if region_tag is None:
            assert len(list(self.job_pending_chunk_ids.keys())) == 1, "Must specify region_tag if there are multiple regions"
            region_tag = list(self.job_pending_chunk_ids.keys())[0]
        if len(self.job_chunk_requests) == 0:
            return None
        bytes_remaining_per_job = {}
        for job_uuid in self.job_pending_chunk_ids.keys():
            bytes_remaining_per_job[job_uuid] = sum(
                [
                    cr.chunk_length_bytes
                    for cr in self.job_chunk_requests[job_uuid].values()
                    if cr.chunk_id in self.job_pending_chunk_ids[job_uuid][region_tag]
                ]
            )
        logger.fs.debug(f"[TransferProgressTracker] Bytes remaining per job: {bytes_remaining_per_job}")
        print(f"[TransferProgressTracker] Bytes remaining per job: {bytes_remaining_per_job}")
        return sum(bytes_remaining_per_job.values())

    def query_bytes_dispatched(self):
        """Query the total number of bytes dispatched to chunks ready for transfer"""
        if len(self.job_chunk_requests) == 0:
            return 0
        bytes_total_per_job = {}
        for job_uuid in self.job_complete_chunk_ids.keys():
            bytes_total_per_job[job_uuid] = sum([cr.chunk_length_bytes for cr in self.job_chunk_requests[job_uuid].values()])
        return sum(bytes_total_per_job.values())



import argparse
from multiprocessing import Manager
from pprint import pprint
import atexit
import json
import os
import signal
import sys
from multiprocessing import Event, Queue
from os import PathLike
from pathlib import Path
from typing import Dict, List

from skyplane.utils import logger

from skyplane.gateway.gateway_queue import GatewayQueue, GatewayANDQueue
from skyplane.gateway.chunk_store import ChunkStore
from skyplane.gateway.gateway_daemon_api import GatewayDaemonAPI
from skyplane.gateway.operators.gateway_operator import (
    GatewaySender,
    GatewayRandomDataGen,
    GatewayWriteLocal,
    GatewayObjStoreReadOperator,
    GatewayObjStoreWriteOperator,
    GatewayWaitReceiver,
)
from skyplane.gateway.operators.gateway_receiver import GatewayReceiver
from skyplane.utils import logger
from collections import defaultdict


# TODO: add default partition ID to main
# create gateway br
class GatewayDaemon:
    def __init__(
        self,
        region: str,
        chunk_dir: PathLike,
        max_incoming_ports=64,
        use_tls=True,
        use_e2ee=True,  # TODO: read from operator field
        use_compression=True,  # TODO: read from operator field
    ):
        # read gateway program
        gateway_program_path = Path(os.environ["GATEWAY_PROGRAM_FILE"]).expanduser()
        gateway_program = json.load(open(gateway_program_path, "r"))

        self.upload_id_map = Manager().dict()

        pprint(gateway_program)

        # read gateway info
        gateway_info_path = Path(os.environ["GATEWAY_INFO_FILE"]).expanduser()
        self.gateway_info = json.load(open(gateway_info_path, "r"))

        print("starting gateway daemon", gateway_program_path)
        pprint(gateway_program)
        assert len(gateway_program) > 0, f"Cannot have empty gateway program {gateway_program}"

        self.use_tls = use_tls

        # todo max_incoming_ports should be configurable rather than static
        self.region = region
        self.max_incoming_ports = max_incoming_ports

        # the chunk store managest the incoming queue of chunks and outgoing queue of chunk status updates
        self.chunk_store = ChunkStore(chunk_dir)

        self.error_event = Event()
        self.error_queue = Queue()
        if use_e2ee:
            e2ee_key_path = Path(os.environ["E2EE_KEY_FILE"]).expanduser()
            with open(e2ee_key_path, "rb") as f:
                self.e2ee_key_bytes = f.read()
            print("Server side E2EE key loaded: ", self.e2ee_key_bytes)
        else:
            self.e2ee_key_bytes = None

        # create gateway operators
        self.terminal_operators = defaultdict(list)  # track terminal operators per partition
        self.num_required_terminal = {}
        self.operators = self.create_gateway_operators(gateway_program)

        # single gateway receiver
        self.gateway_receiver = GatewayReceiver(
            "reciever",
            region=region,
            chunk_store=self.chunk_store,
            error_event=self.error_event,
            error_queue=self.error_queue,
            max_pending_chunks=max_incoming_ports,
            use_tls=self.use_tls,
            use_compression=use_compression,
            e2ee_key_bytes=self.e2ee_key_bytes,
        )

        # API server
        self.api_server = GatewayDaemonAPI(
            self.chunk_store,
            self.gateway_receiver,
            self.error_event,
            self.error_queue,
            terminal_operators=self.terminal_operators,
            num_required_terminal=self.num_required_terminal,
            upload_id_map=self.upload_id_map,
        )
        self.api_server.start()
        atexit.register(self.api_server.shutdown)
        logger.info(f"[gateway_daemon] API started at {self.api_server.url}")

    def print_operator_graph(self):
        def print_operator_graph_helper(partition, queue: GatewayQueue, prefix: str):
            if queue is None:
                return
            print(f"{prefix} {partition}: Input queue:", queue, "handles:", queue.get_handles())
            for handle in queue.get_handles():
                print(f"{prefix} {partition}: Operator {handle}")
                # TODO: causes error sometimes for mux_or
                next_queue = self.operators[handle].output_queue
                print_operator_graph_helper(partition, next_queue, prefix + "--")

        for partition, queue in self.chunk_store.chunk_requests.items():
            print(f"Partition {partition}, {queue}")
            print_operator_graph_helper(partition, queue, "")

    def create_gateway_operators(self, gateway_program: Dict):
        """Create a gateway plan from a gateway program"""

        operators = {}

        def create_output_queue(operator: Dict):
            # create output data queue
            if len(operator["children"]) == 0:
                return None
            print("get output queue", operator["op_type"], operator["children"][0]["op_type"])
            if operator["children"][0]["op_type"] == "mux_and":
                return GatewayANDQueue()
            return GatewayQueue()

        def get_child_operators(operator):
            if len(operator["children"]) == 0:
                return []
            if operator["children"][0]["op_type"] == "mux_and" or operator["children"][0]["op_type"] == "mux_or":
                return operator["children"][0]["children"]
            return operator["children"]

        def create_gateway_operators_helper(input_queue, program: List[Dict], partition_ids: List[str]):
            total_p = 0
            for op in program:
                handle = op["op_type"] + "_" + op["handle"]
                print(f"Input queue {input_queue}, adding handle {handle} (current handles {input_queue.get_handles()}")
                input_queue.register_handle(handle)

                # get child operators
                child_operators = get_child_operators(op)

                if op["op_type"] == "mux_or":
                    # parent must have been mux_and
                    assert isinstance(
                        input_queue, GatewayANDQueue
                    ), f"Parent must have been mux_and {handle}, instead was {input_queue} {gateway_program}"

                    # recurse to children with single queue
                    total_p += create_gateway_operators_helper(input_queue.get_handle_queue(handle), child_operators, partition_ids)
                    continue

                # create output data queue
                output_queue = create_output_queue(op)
                if isinstance(output_queue, GatewayANDQueue):
                    # update number of operations that must be completed
                    for partition in partition_ids:
                        self.num_required_terminal[str(partition)] = self.num_required_terminal[str(partition)] - 1 + len(child_operators)

                print(f"Input queue {input_queue} handle {handle}: created output queue {output_queue}")
                print(f"Input queue {input_queue} handle {handle}: has children {child_operators}")
                if output_queue is None:
                    # track what opeartors need to complete processing the chunk
                    for partition in partition_ids:
                        self.terminal_operators[str(partition)].append(handle)

                # create operators
                if op["op_type"] == "receive":
                    # wait for chunks from receiver
                    operators[handle] = GatewayWaitReceiver(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        n_processes=1,  # dummy wait thread, not actual receiver
                        chunk_store=self.chunk_store,
                        error_event=self.error_event,
                        error_queue=self.error_queue,
                    )
                    total_p += 1
                elif op["op_type"] == "read_object_store":
                    operators[handle] = GatewayObjStoreReadOperator(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        n_processes=op["num_connections"],
                        chunk_store=self.chunk_store,
                        bucket_name=op["bucket_name"],
                        bucket_region=op["bucket_region"],
                    )
                    total_p += op["num_connections"]
                elif op["op_type"] == "gen_data":
                    operators[handle] = GatewayRandomDataGen(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        chunk_store=self.chunk_store,
                        size_mb=op["size_mb"],
                    )
                elif op["op_type"] == "send":
                    # TODO: handle private ips for GCP->GCP
                    target_gateway_info = self.gateway_info[op["target_gateway_id"]]
                    ip_addr = target_gateway_info["private_ip_address"] if op["private_ip"] else target_gateway_info["public_ip_address"]
                    print("Gateway sender sending to ", ip_addr, "private", op["private_ip"])
                    operators[handle] = GatewaySender(
                        handle,
                        region=self.region,
                        ip_addr=ip_addr,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_event=self.error_event,
                        error_queue=self.error_queue,
                        chunk_store=self.chunk_store,
                        use_tls=self.use_tls,
                        use_compression=op["compress"],
                        e2ee_key_bytes=self.e2ee_key_bytes,
                        n_processes=op["num_connections"],
                    )
                    total_p += op["num_connections"]
                elif op["op_type"] == "write_object_store":
                    operators[handle] = GatewayObjStoreWriteOperator(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        n_processes=op["num_connections"],
                        chunk_store=self.chunk_store,
                        bucket_name=op["bucket_name"],
                        bucket_region=op["bucket_region"],
                        upload_id_map=self.upload_id_map,
                        prefix=op["key_prefix"],
                    )
                    total_p += op["num_connections"]
                elif op["op_type"] == "write_local":
                    operators[handle] = GatewayWriteLocal(
                        handle=handle,
                        region=self.region,
                        input_queue=input_queue,
                        output_queue=output_queue,
                        error_queue=self.error_queue,
                        error_event=self.error_event,
                        chunk_store=self.chunk_store,
                    )
                    total_p += 1
                else:
                    raise ValueError(f"Unsupported op_type {op['op_type']}")
                # recursively create for child operators
                total_p += create_gateway_operators_helper(output_queue, child_operators, partition_ids)
            return total_p

        pprint(gateway_program)

        # create operator tree for each partition
        total_p = 0
        for program_group in gateway_program:
            partitions = program_group["partitions"]
            program = program_group["value"]

            print("partitions", partitions)
            # create initial queue for partition
            if program[0]["op_type"] == "mux_and":
                queue = GatewayANDQueue()
                print(f"First operator is mux_and: queue {queue}")
                assert len(program) == 1, f"mux_and cannot have siblings"
                program = program[0]["children"]
                for partition in partitions:
                    self.num_required_terminal[str(partition)] = len(program)
            else:
                queue = GatewayQueue()
                print(f"First operator is ", program[0], "queue", queue)

                for partition in partitions:
                    self.num_required_terminal[str(partition)] = 1

            # link all partitions to same queue reference
            for partition in partitions:
                self.chunk_store.add_partition(str(partition), queue)

            # create DAG for this partition group
            total_p += create_gateway_operators_helper(
                self.chunk_store.chunk_requests[str(partition)],  # incoming chunk requests for partition
                program,  # single partition program
                partitions,
            )
        # print("TOTAL NUMBER OF PROCESSES", total_p)
        return operators

    def run(self):
        exit_flag = Event()

        def exit_handler(signum, frame):
            logger.warning("[gateway_daemon] Received signal {}. Exiting...".format(signum))
            exit_flag.set()
            for operator in self.operators.values():
                operator.stop_workers()
            sys.exit(0)

        for operator in self.operators.values():
            logger.info(f"[gateway_daemon] Starting gateway operator {operator.handle} workers")
            operator.start_workers()

        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGTERM, exit_handler)

        logger.info("[gateway_daemon] Starting daemon loop")
        try:
            while not exit_flag.is_set() and not self.error_event.is_set():
                self.api_server.pull_chunk_status_queue()
        except Exception as e:
            self.error_queue.put(e)
            self.error_event.set()
            logger.error(f"[gateway_daemon] Exception in daemon loop: {e}")
            logger.exception(e)

        # shut down workers except for API to report status
        logger.info("[gateway_daemon] Exiting all workers except for API")
        for operator in self.operators.values():
            operator.stop_workers()
        logger.info("[gateway_daemon] Done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Skyplane Gateway Daemon")
    parser.add_argument("--region", type=str, required=True, help="Region tag (provider:region")
    parser.add_argument("--chunk-dir", type=Path, default="/tmp/skyplane/chunks", help="Directory to store chunks")
    parser.add_argument("--disable-tls", action="store_true")
    parser.add_argument("--use-compression", action="store_true")  # TODO: remove
    parser.add_argument("--disable-e2ee", action="store_true")  # TODO: remove
    args = parser.parse_args()

    os.makedirs(args.chunk_dir)
    daemon = GatewayDaemon(
        region=args.region,
        chunk_dir=args.chunk_dir,
        use_tls=not args.disable_tls,
    )
    daemon.run()



from functools import partial
import sys
from typing import Optional

from rich.console import Console
from rich import print as rprint

import typer
import traceback as tb

from skyplane import compute
from skyplane.utils import logger
from skyplane.utils.fn import do_parallel

console = Console()


def print_header():
    header = """ _____ _   ____   _______ _       ___   _   _  _____ 
/  ___| | / /\ \ / / ___ \ |     / _ \ | \ | ||  ___|
\ `--.| |/ /  \ V /| |_/ / |    / /_\ \|  \| || |__  
 `--. \    \   \ / |  __/| |    |  _  || . ` ||  __| 
/\__/ / |\  \  | | | |   | |____| | | || |\  || |___ 
\____/\_| \_/  \_/ \_|   \_____/\_| |_/\_| \_/\____/"""
    console.print(f"[bright_black]{header}[/bright_black]\n")


def print_stats_completed(total_runtime_s: float, throughput_gbits: Optional[float]):
    console.print(f"\n:white_check_mark: [bold green]Transfer completed successfully[/bold green]")
    runtime_line = f"[white]Transfer runtime:[/white] [bright_black]{total_runtime_s:.2f}s[/bright_black]"
    throughput_line = (
        f", [white]Throughput:[/white] [bright_black]{throughput_gbits:.2f}Gbps[/bright_black]" if throughput_gbits is not None else ""
    )
    console.print(f"{runtime_line}{throughput_line}")


def query_instances():
    instances = []
    query_jobs = []

    def catch_error(fn):
        def run():
            try:
                return fn()
            except Exception as e:
                logger.error(f"Error encountered during deprovision: {e}")
                return []

        return run

    if compute.AWSAuthentication().enabled():
        aws = compute.AWSCloudProvider()
        for region in aws.region_list():
            query_jobs.append(catch_error(partial(aws.get_matching_instances, region)))
    if compute.AzureAuthentication().enabled():
        query_jobs.append(catch_error(lambda: compute.AzureCloudProvider().get_matching_instances()))
    if compute.GCPAuthentication().enabled():
        query_jobs.append(catch_error(lambda: compute.GCPCloudProvider().get_matching_instances()))
    # query in parallel
    for instance_list in do_parallel(
        lambda f: f(), query_jobs, n=-1, return_args=False, spinner=True, desc="Querying clouds for instances"
    ):
        instances.extend(instance_list)
    return instances


# pretty exception handler with rich
def register_exception_handler():
    def exception_handler(exception_type, exception, traceback, debug_hook=sys.excepthook):
        # write full traceback infomation with locals to log file
        logger.fs.error(f"Uncaught exception: {exception_type.__name__}: {exception}")
        logger.fs.error("Traceback:\n" + "".join(tb.format_exception(exception_type, exception, traceback)))
        rprint(f"[red][bold]Uncaught exception:[/bold] ({exception_type.__name__}) {exception}[/red]", file=sys.stderr)
        typer.secho(
            "Please check the log file for more information, and ensure to include it if reporting an issue on Github.",
            fg=typer.colors.YELLOW,
        )
        sys.exit(1)

    sys.excepthook = exception_handler
